/*

  P-Code interpreter (to run the apple pascal system)
  Copyright (C) 2000 Mario Klebsch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdio.h>

static const char *Instructions[256]= {
  "sldc	0", 
  "sldc	1", 
  "sldc	2", 
  "sldc	3", 
  "sldc	4", 
  "sldc	5", 
  "sldc	6", 
  "sldc	7", 
  "sldc	8", 
  "sldc	9", 
  "sldc	10", 
  "sldc	11", 
  "sldc	12", 
  "sldc	13", 
  "sldc	14", 
  "sldc	15", 
  "sldc	16", 
  "sldc	17", 
  "sldc	18", 
  "sldc	19", 
  "sldc	20", 
  "sldc	21", 
  "sldc	22", 
  "sldc	23", 
  "sldc	24", 
  "sldc	25", 
  "sldc	26", 
  "sldc	27", 
  "sldc	28", 
  "sldc	29", 
  "sldc	30", 
  "sldc	31", 
  "sldc	32", 
  "sldc	33", 
  "sldc	34", 
  "sldc	35", 
  "sldc	36", 
  "sldc	37", 
  "sldc	38", 
  "sldc	39", 
  "sldc	40", 
  "sldc	41", 
  "sldc	42", 
  "sldc	43", 
  "sldc	44", 
  "sldc	45", 
  "sldc	46", 
  "sldc	47", 
  "sldc	48", 
  "sldc	49", 
  "sldc	50", 
  "sldc	51", 
  "sldc	52", 
  "sldc	53", 
  "sldc	54", 
  "sldc	55", 
  "sldc	56", 
  "sldc	57", 
  "sldc	58", 
  "sldc	59", 
  "sldc	60", 
  "sldc	61", 
  "sldc	62", 
  "sldc	63", 
  "sldc	64", 
  "sldc	65", 
  "sldc	66", 
  "sldc	67", 
  "sldc	68", 
  "sldc	69", 
  "sldc	70", 
  "sldc	71", 
  "sldc	72", 
  "sldc	73", 
  "sldc	74", 
  "sldc	75", 
  "sldc	76", 
  "sldc	77", 
  "sldc	78", 
  "sldc	79", 
  "sldc	80", 
  "sldc	81", 
  "sldc	82", 
  "sldc	83", 
  "sldc	84", 
  "sldc	85", 
  "sldc	86", 
  "sldc	87", 
  "sldc	88", 
  "sldc	89", 
  "sldc	90", 
  "sldc	91", 
  "sldc	92", 
  "sldc	93", 
  "sldc	94", 
  "sldc	95", 
  "sldc	96", 
  "sldc	97", 
  "sldc	98", 
  "sldc	99", 
  "sldc	100", 
  "sldc	101", 
  "sldc	102", 
  "sldc	103", 
  "sldc	104", 
  "sldc	105", 
  "sldc	106", 
  "sldc	107", 
  "sldc	108", 
  "sldc	109", 
  "sldc	110", 
  "sldc	111", 
  "sldc	112", 
  "sldc	113", 
  "sldc	114", 
  "sldc	115", 
  "sldc	116", 
  "sldc	117", 
  "sldc	118", 
  "sldc	119", 
  "sldc	120", 
  "sldc	121", 
  "sldc	122", 
  "sldc	123", 
  "sldc	124", 
  "sldc	125", 
  "sldc	126", 
  "sldc	127", 
  "abi		(absolute value integer)", 
  "abr		(absolute value real)", 
  "adi		(add integer)", 
  "adr		(add real)", 
  "land		(logical and)", 
  "dif		(set difference)", 
  "dvi		(divide integer)", 
  "dvr		(divide real)", 
  "chk		(check)", 
  "flo		(float tos-1)", 
  "flt		(float tos)", 
  "inn		(set memberschip)", 
  "int		(set intersection)", 
  "lor		(logical or)", 
  "modi		(modulo integer)", 
  "mpi		(multiply integer)", 
  "mpr		(multiply real)", 
  "ngi		(negate integer)", 
  "ngr		(negate real)", 
  "lnot		(logical not)", 
  "srs		(build subrange set)", 
  "sbi		(substract integer)", 
  "sbr		(substract real)", 
  "sgs		(build singleton set)", 
  "sqi		(square integer)", 
  "sqr		(square real)", 
  "sto		(store indirect word)", 
  "ixs		(index string array)", 
  "uni		(set union)", 
  "lde	U,B	(load extended word)", 
  "csp	Q	(call standard procedure)", 
  "ldcn		(load constant nil)", 
  "adj	U	(adjust set)", 
  "fjp	S	(false jump)", 
  "inc	B	(increment field pointer)", 
  "ind	B	(load indirect word)", 
  "ixa	B	(index array)", 
  "lao	B	(load global address)", 
  "lsa	Y	(load string address)", 
  "lae	U,B	(load extended address)", 
  "mov	B	(move words)", 
  "ldo	B	(load global word)", 
  "sas	U	(string assign)", 
  "sro	B	(store global word)", 
  "xjp	W,W,W,V	(case jump)", 
  "rnp	D	(return non-base procedure)", 
  "cip	U	(call intermediate procedure)", 
  "equ	T	(equal)", 
  "geq	T	(greater or equal)", 
  "grt	T	(greater)", 
  "lda	D,B	(load intermediate address)", 
  "ldc	X	(load multiple word constant)", 
  "leq	T	(less or equal)", 
  "les	T	(less than)", 
  "lod	D,B	(load intermediate word)", 
  "neq	T	(not equal)", 
  "str	D,B	(store intermediate word)", 
  "ujp	S	(unconditional jump)", 
  "ldp		(load packed field)", 
  "stp		(store into packed field)", 
  "ldm	U	(load multiple words)", 
  "stm	U	(store multiple words)", 
  "ldb		(load byte)", 
  "stb		(store byte)", 
  "ixp	U,U	(index packed array)", 
  "rbp	D	(return base procedure)", 
  "cbp	U	(call base procedure)", 
  "equi		(equal integer)", 
  "geqi		(greater or equal integer)", 
  "grti		(greater iteger)", 
  "lla	B	(load local address)", 
  "ldci	W	(load constant integer)", 
  "leqi		(less or equal integer)", 
  "lesi		(less than integer)", 
  "ldl	B	(load local word)", 
  "neqi		(not equal integer)", 
  "stl	B	(store local word)", 
  "cxp	U,U	(call external procedure)", 
  "clp	U,	(call local procedure)", 
  "cgp	U	(call global procedure)", 
  "lpa	Z	(load packed array)", 
  "ste	U,B	(store extended word)", 
  ".db	210", 
  "efj	S	(equal false jump)", 
  "nfj	S	(not equal false jump)", 
  "bpt		(breakpoint)", 
  "xit		(exit operating system)", 
  "nop		(no operation)", 
  "sldl	1	(short load local word)", 
  "sldl	2	(short load local word)", 
  "sldl	3	(short load local word)", 
  "sldl	4	(short load local word)", 
  "sldl	5	(short load local word)", 
  "sldl	6	(short load local word)", 
  "sldl	7	(short load local word)", 
  "sldl	8	(short load local word)", 
  "sldl	9	(short load local word)", 
  "sldl	10	(short load local word)", 
  "sldl	11	(short load local word)", 
  "sldl	12	(short load local word)", 
  "sldl	13	(short load local word)", 
  "sldl	14	(short load local word)", 
  "sldl	15	(short load local word)", 
  "sldl	16	(short load local word)", 
  "sldo	1	(short load global word)", 
  "sldo	2	(short load global word)", 
  "sldo	3	(short load global word)", 
  "sldo	4	(short load global word)", 
  "sldo	5	(short load global word)", 
  "sldo	6	(short load global word)", 
  "sldo	7	(short load global word)", 
  "sldo	8	(short load global word)", 
  "sldo	9	(short load global word)", 
  "sldo	10	(short load global word)", 
  "sldo	11	(short load global word)", 
  "sldo	12	(short load global word)", 
  "sldo	13	(short load global word)", 
  "sldo	14	(short load global word)", 
  "sldo	15	(short load global word)", 
  "sldo	16	(short load global word)", 
  "sind	0	(short load indirect word)", 
  "sind	1	(short load indirect word)", 
  "sind	2	(short load indirect word)", 
  "sind	3	(short load indirect word)", 
  "sind	4	(short load indirect word)", 
  "sind	5	(short load indirect word)", 
  "sind	6	(short load indirect word)", 
  "sind	7	(short load indirect word)"
};

int DisasmP(char *Buffer, int Addr)
{
  unsigned chat OpCode=Apple_memory_read(Addr++);
  char		*s=Instructions[OpCode];
  char		*d=Buffer;
  char		ch;
  int		Val;

  while ((ch=*s++))
    switch(ch)
      {
      case 'B':
	Val=Apple_memory_read(Addr++);
	if (Val&0x80)
	  Val=((Val&0x7f)<<8)+Apple_memory_read(Addr++);
	sprintf(d,"%d", Val);
	d +=strlen(d);
	break;
      case 'D':
      case 'U':
	Val=Apple_memory_read(Addr++);
	sprintf(d,"%d", Val);
	d +=strlen(d);
	break;
      case 'S':
	Val=Apple_memory_read(Addr++);
	if (Val&0x80)
	  Val=0x100-Val;
	sprintf(d,"%d", Val);
	d +=strlen(d);
	break;
      case 'W':
	Val=Apple_memory_read(Addr++);
	Val |= (Apple_memory_read(Addr++))<<8;
	sprintf(d,"%d", Val);
	d +=strlen(d);
	break;
      default:
	*d++=ch;
      }
  *d++='\0';
}
